from typing import Literal

import click
import typer
from click import Choice
from rich import print

from .collision_event import CollisionEvent
from .constants import (
    AFTER_SIMULATION_TEXT,
    CAR_COMMANDS_PROMPT_TEXT,
    CAR_NAME_PROMPT_TEXT,
    CAR_POSITION_PROMPT_TEXT,
    CURRENT_CAR_LIST_TEXT,
    FIELD_CREATION_TEXT,
    FIELD_DIMS_PROMPT_TEXT,
    INPUT_PROMPT_SUFFIX,
    INTRO_TEXT,
    MENU_PROMPT_SUFFIX,
    POST_SIMULATION_OPTIONS_MENU_TEXT,
    PRE_SIMULATION_OPTIONS_MENU_TEXT,
    PRESS_ANY_KEY_TO_CONTINUE_SIMULATION_TEXT,
    SIMULATION_COMPLETED_TEXT,
)
from .custom_paramtypes import (
    CarNameParamType,
    CarPositionVectorParamType,
    CustomCarCommandParamType,
    CustomFieldDimsParamType,
    PreSimulationSelectionParamType,
)
from .field import Field
from .simulator import Simulator
from .state import State
from .state_renderer import render_state_on_field


def render_intro_text():
    print(INTRO_TEXT)


def prompt_field_dims_and_initialize_field() -> Field:
    field_dims = typer.prompt(
        FIELD_DIMS_PROMPT_TEXT,
        type=CustomFieldDimsParamType,
        prompt_suffix=INPUT_PROMPT_SUFFIX,
    )

    field = Field(*field_dims)
    print(FIELD_CREATION_TEXT.format(field.width, field.height))

    return field


def prompt_pre_simulation_menu_and_get_selection(
    simulator: Simulator,
) -> Literal["1", "2", "3"]:
    return typer.prompt(
        PRE_SIMULATION_OPTIONS_MENU_TEXT,
        prompt_suffix=MENU_PROMPT_SUFFIX,
        type=PreSimulationSelectionParamType(simulator),
    )


def print_initial_car_positions_and_commands_in_simulator(simulator: Simulator):
    print(CURRENT_CAR_LIST_TEXT)

    init_car_position_map = simulator.init_car_position_map
    init_car_command_map = simulator.init_car_command_map

    cars = sorted(init_car_position_map.keys())
    for car in cars:
        position_vector = init_car_position_map[car]
        position = position_vector.position
        direction = position_vector.direction
        command_queue = init_car_command_map[car]
        print(
            f"- {car}, {position} {direction}, {''.join(map(lambda _cmd: _cmd.value, command_queue)) or '[]'}"
        )


def prompt_car_name_position_and_commands(simulator: Simulator) -> None:
    car = typer.prompt(
        CAR_NAME_PROMPT_TEXT,
        prompt_suffix=INPUT_PROMPT_SUFFIX,
        type=CarNameParamType(simulator),
    )

    position_vector = typer.prompt(
        CAR_POSITION_PROMPT_TEXT.format(car),
        prompt_suffix=INPUT_PROMPT_SUFFIX,
        type=CarPositionVectorParamType(simulator.field),
    )

    command_queue = typer.prompt(
        CAR_COMMANDS_PROMPT_TEXT.format(car),
        prompt_suffix=INPUT_PROMPT_SUFFIX,
        type=CustomCarCommandParamType,
        default="",
        show_default=False,
    )

    simulator.add_car(car, position_vector, command_queue)
    print_initial_car_positions_and_commands_in_simulator(simulator)


def print_car_positions_in_state(state: State) -> None:

    car_position_map = state.car_position_map

    cars = sorted(car_position_map.keys())
    for car in cars:
        position_vector = car_position_map[car]
        position = position_vector.position
        direction = position_vector.direction
        print(f"- {car}, {position} {direction}")


def print_collision_history(collision_history: list[CollisionEvent]) -> None:
    for collision_event in collision_history:
        print(f"- {collision_event.collision_report}")


def render_simulation_state(
    simulator: Simulator, state: State, collision_history: list[CollisionEvent]
) -> None:
    render_state_on_field(state, simulator.field)
    print_car_positions_in_state(state)
    print_collision_history(collision_history)


def render_simulation_outcome(simulator: Simulator) -> None:
    print_initial_car_positions_and_commands_in_simulator(simulator)

    initial_state, collision_history = (
        simulator.get_initial_state_and_collision_history()
    )
    render_simulation_state(simulator, initial_state, collision_history)

    print(AFTER_SIMULATION_TEXT)

    final_state, collision_history = simulator.run_simulation()
    render_simulation_state(simulator, final_state, collision_history)


def render_interactive_simulation(simulator: Simulator):
    print_initial_car_positions_and_commands_in_simulator(simulator)

    for state, collision_history in simulator.run_simulation_generator():
        render_simulation_state(simulator, state, collision_history)
        click.pause(PRESS_ANY_KEY_TO_CONTINUE_SIMULATION_TEXT)

    print(SIMULATION_COMPLETED_TEXT)


def prompt_post_simulation_menu_and_get_selection() -> Literal["1", "2"]:
    return typer.prompt(
        POST_SIMULATION_OPTIONS_MENU_TEXT,
        prompt_suffix=MENU_PROMPT_SUFFIX,
        type=Choice(["1", "2"], case_sensitive=False),
        show_choices=False,
    )
