from dataclasses import dataclass
from typing import Optional, Self

from .collision_rubble import CollisionRubble
from .enums import CollisionType
from .position import Position
from .state import State


@dataclass
class CollisionEvent:

    car: str
    state: State
    position: Position
    collision_type: CollisionType
    collision_rubble: CollisionRubble
    collision_report: str

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CollisionEvent):
            return NotImplemented

        return (
            self.car == other.car
            and self.state == other.state
            and self.position == other.position
            and self.collision_type == other.collision_type
            and self.collision_rubble == other.collision_rubble
            and self.collision_report == other.collision_report
        )

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, CollisionEvent):
            return NotImplemented

        if self.state != other.state:
            return self.state.id < other.state.id

        return self.car < other.car

    @staticmethod
    def get_collision_report_for_collision_at_outset(
        car: str, position: Position, collision_rubble: CollisionRubble
    ) -> str:
        return f"{car} in {collision_rubble} because of these cars being placed together at {position} at simulation outset (step 0)"

    @classmethod
    def from_collision_at_outset(
        cls,
        car: str,
        state: State,
        position: Position,
        collision_rubble: CollisionRubble,
    ) -> Self:
        collision_report = CollisionEvent.get_collision_report_for_collision_at_outset(
            car, position, collision_rubble
        )

        return cls(
            car=car,
            state=state,
            collision_type=CollisionType.COLLISION_AT_OUTSET,
            position=position,
            collision_rubble=collision_rubble,
            collision_report=collision_report,
        )

    @staticmethod
    def get_collision_report_for_collision_at_empty_position(
        car: str, state: State, position: Position, collision_rubble: CollisionRubble
    ) -> str:
        remaining_cars = list(filter(lambda _car: _car != car, collision_rubble.cars))
        return f"{car}, collides with {', '.join(remaining_cars)} at {position} at step {state.id}"

    @classmethod
    def from_collision_at_empty_position(
        cls,
        car: str,
        state: State,
        position: Position,
        collision_rubble: CollisionRubble,
    ) -> Self:

        collision_report = (
            CollisionEvent.get_collision_report_for_collision_at_empty_position(
                car, state, position, collision_rubble
            )
        )

        return cls(
            car=car,
            state=state,
            collision_type=CollisionType.COLLISION_AT_EMPTY_POSITION,
            position=position,
            collision_rubble=collision_rubble,
            collision_report=collision_report,
        )

    @staticmethod
    def get_collision_report_for_collision_with_stationary_car(
        car: str, state: State, position: Position, stationary_car: str
    ) -> str:
        return f"{car}, collides with stationary {stationary_car} at {position} at step {state.id}"

    @classmethod
    def from_collision_with_stationary_car(
        cls,
        car: str,
        state: State,
        position: Position,
        stationary_car: str,
        collision_rubble: CollisionRubble,
    ) -> Self:

        collision_report = (
            CollisionEvent.get_collision_report_for_collision_with_stationary_car(
                car, state, position, stationary_car
            )
        )

        return cls(
            car=car,
            state=state,
            collision_type=CollisionType.COLLISION_WITH_STATIONARY_CAR,
            position=position,
            collision_rubble=collision_rubble,
            collision_report=collision_report,
        )

    @staticmethod
    def get_collision_report_for_collision_with_rubble(
        car: str,
        state: State,
        position: Position,
        prev_collision_rubble: CollisionRubble,
    ) -> str:
        return f"{car}, collides with pre-existing {prev_collision_rubble} at {position} at step {state.id}"

    @classmethod
    def from_collision_with_rubble(
        cls,
        car: str,
        state: State,
        position: Position,
        prev_collision_rubble: CollisionRubble,
        collision_rubble: CollisionRubble,
    ) -> Self:

        collision_report = (
            CollisionEvent.get_collision_report_for_collision_with_rubble(
                car, state, position, prev_collision_rubble
            )
        )

        return cls(
            car=car,
            state=state,
            collision_type=CollisionType.COLLISION_WITH_PRE_EXISTING_RUBBLE,
            position=position,
            collision_rubble=collision_rubble,
            collision_report=collision_report,
        )


def generate_collision_history_at_simulation_onset(
    state: State,
) -> list[CollisionEvent]:
    collision_history: list[CollisionEvent] = []

    for position, collision_rubble in state.collision_rubble_map.items():
        for car in collision_rubble.cars:
            collision_event = CollisionEvent.from_collision_at_outset(
                car=car,
                state=state,
                position=position,
                collision_rubble=collision_rubble,
            )
            collision_history.append(collision_event)

    return collision_history


def generate_collision_history_for_collisions_with_rubble(
    prev_state: State, cur_state: State
):
    collision_history: list[CollisionEvent] = []

    for position, collision_rubble in cur_state.collision_rubble_map.items():
        if position not in prev_state.collision_rubble_map:
            continue

        prev_collision_rubble = prev_state.collision_rubble_map[position]

        for car in collision_rubble.cars:
            if car in prev_collision_rubble.cars:
                continue

            collision_event = CollisionEvent.from_collision_with_rubble(
                car,
                cur_state,
                position,
                prev_collision_rubble,
                collision_rubble,
            )
            collision_history.append(collision_event)

    return collision_history


def get_stationary_car_in_collision_rubble(
    collison_rubble: CollisionRubble,
    position: Position,
    prev_state: State,
) -> Optional[str]:

    for car in collison_rubble.cars:
        prev_car_position = prev_state.car_position_map[car].position

        if prev_car_position == position:
            return car

    return None


def generate_collision_history_for_collisions_with_stationary_car(
    state: State,
    position: Position,
    stationary_car: str,
    collision_rubble: CollisionRubble,
) -> list[CollisionEvent]:
    collision_history: list[CollisionEvent] = []

    for car in collision_rubble.cars:
        if car != stationary_car:
            collision_event = CollisionEvent.from_collision_with_stationary_car(
                car,
                state,
                position,
                stationary_car,
                collision_rubble,
            )
            collision_history.append(collision_event)

    return collision_history


def generate_collision_history_for_collisions_at_empty_positions(
    state: State, position: Position, collision_rubble: CollisionRubble
) -> list[CollisionEvent]:
    collision_history: list[CollisionEvent] = []

    for car in collision_rubble.cars:
        collision_event = CollisionEvent.from_collision_at_empty_position(
            car,
            state,
            position,
            collision_rubble,
        )
        collision_history.append(collision_event)

    return collision_history


def generate_collision_history_for_mutual_car_collisions(
    prev_state: State, cur_state: State
) -> list[CollisionEvent]:
    collision_history: list[CollisionEvent] = []

    rubble_collision_history: list[CollisionEvent]
    for position, collision_rubble in cur_state.collision_rubble_map.items():

        if position in prev_state.collision_rubble_map:
            continue

        stationary_car = get_stationary_car_in_collision_rubble(
            collision_rubble, position, prev_state
        )

        if stationary_car:
            rubble_collision_history = (
                generate_collision_history_for_collisions_with_stationary_car(
                    cur_state, position, stationary_car, collision_rubble
                )
            )

        else:
            rubble_collision_history = (
                generate_collision_history_for_collisions_at_empty_positions(
                    cur_state, position, collision_rubble
                )
            )

        collision_history.extend(rubble_collision_history)

    return collision_history


def generate_collision_history_for_current_state(
    prev_state: Optional[State], cur_state: State
) -> list[CollisionEvent]:

    collision_history: list[CollisionEvent]

    if prev_state is None:
        collision_history = generate_collision_history_at_simulation_onset(cur_state)

    else:
        collision_history_rubble = (
            generate_collision_history_for_collisions_with_rubble(prev_state, cur_state)
        )
        collision_history_mutual = generate_collision_history_for_mutual_car_collisions(
            prev_state, cur_state
        )

        collision_history = collision_history_rubble + collision_history_mutual

    return sorted(collision_history, key=lambda collision_event: collision_event.car)
