from collections import deque
from dataclasses import dataclass
from typing import Any, Literal, Optional

from click.core import Context, Parameter
from click.types import ParamType, StringParamType

from .enums import Command, Direction
from .field import Field
from .helper import get_command_queue_from_string
from .position import Position
from .position_vector import PositionVector
from .simulator import Simulator


class FieldDimsParamType(ParamType):
    name = "field_dimenions"

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> tuple[int, int]:
        splits = value.split(self.envvar_list_splitter)

        if len(splits) != 2:
            if len(splits) == 1:
                self.fail(
                    f"2 values are required, but only 1 was given.",
                    param,
                    ctx,
                )
            else:
                self.fail(
                    f"2 values are required, but {len(splits)} were given.",
                    param,
                    ctx,
                )

        if not splits[0].isdigit():
            self.fail(f"{splits[0]} is not a valid integer", param, ctx)

        if not splits[1].isdigit():
            self.fail(f"{splits[1]} is not a valid integer", param, ctx)

        return int(splits[0]), int(splits[1])


CustomFieldDimsParamType = FieldDimsParamType()


@dataclass
class CarNameParamType(StringParamType):
    simulator: Simulator

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> str:
        parsed_value = super().convert(value, param, ctx)

        if self.simulator.is_car_present_in_simulator(parsed_value):
            self.fail(f"car {parsed_value} has already been added to the simulator")

        return parsed_value


@dataclass
class CarPositionVectorParamType(ParamType):
    name = "car_position_vector"
    field: Field

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> PositionVector:
        splits = value.split(self.envvar_list_splitter)

        if len(splits) != 3:
            if len(splits) == 1:
                self.fail(
                    f"3 values are required, but only 1 was given.",
                    param,
                    ctx,
                )
            else:
                self.fail(
                    f"3 values are required, but {len(splits)} were given.",
                    param,
                    ctx,
                )

        if not splits[0].isdigit():
            self.fail(f"{splits[0]} is not a valid integer", param, ctx)

        if not splits[1].isdigit():
            self.fail(f"{splits[1]} is not a valid integer", param, ctx)

        position = Position(int(splits[0]), int(splits[1]))
        if not position.is_within_field_bounds(self.field):
            self.fail(f"{position} is not within {self.field} bounds")

        direction = splits[2]
        if direction not in Direction.__members__.values():
            self.fail(
                f"{direction} is not one of {list(map(lambda enum: enum.value, Direction.__members__.values()))}"
            )

        return PositionVector(position, Direction(direction))


class CarCommandParamType(StringParamType):

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> deque[Command]:
        parsed_value = super().convert(value, param, ctx)

        if not parsed_value:
            return deque()

        for char in parsed_value:
            if char not in Command.__members__.values():
                self.fail(
                    f"{char} is not one of {list(map(lambda enum: enum.value, Command.__members__.values()))}"
                )

        return get_command_queue_from_string(parsed_value)


CustomCarCommandParamType = CarCommandParamType()


@dataclass
class PreSimulationSelectionParamType(StringParamType):
    simulator: Simulator

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Literal["1", "2", "3"]:
        parsed_value = super().convert(value, param, ctx)

        if parsed_value not in ["1", "2", "3"]:
            self.fail(f"{parsed_value} is not of of [1, 2, 3]")

        if parsed_value in ["2", "3"] and self.simulator.is_empty():
            self.fail("cannot run simulation without any cars")

        return parsed_value
