import dataclasses
from collections import deque
from typing import Iterator

from .collision_event import (
    CollisionEvent,
    generate_collision_history_for_current_state,
)
from .enums import Command
from .field import Field
from .position_vector import PositionVector
from .state import State


@dataclasses.dataclass
class Simulator:
    field: Field

    init_car_position_map: dict[str, PositionVector] = dataclasses.field(
        default_factory=lambda: {}
    )
    init_car_command_map: dict[str, deque[Command]] = dataclasses.field(
        default_factory=lambda: {}
    )

    def is_empty(self) -> bool:
        return len(self.init_car_position_map) == 0

    def is_car_present_in_simulator(self, name: str) -> bool:
        return name in self.init_car_position_map

    def add_car(
        self, name: str, position_vector: PositionVector, command_queue: deque[Command]
    ) -> None:
        self.init_car_position_map[name] = position_vector
        self.init_car_command_map[name] = command_queue

    def get_initial_state_and_collision_history(
        self,
    ) -> tuple[State, list[CollisionEvent]]:
        collision_history: list[CollisionEvent] = []

        initial_state = State.from_initial_simulator_config(
            self.init_car_position_map, self.init_car_command_map
        )
        collision_history.extend(
            generate_collision_history_for_current_state(None, initial_state)
        )

        return initial_state, collision_history

    def run_simulation(self) -> tuple[State, list[CollisionEvent]]:

        initial_state, collision_history = (
            self.get_initial_state_and_collision_history()
        )

        prev_state: State
        cur_state: State = initial_state
        while cur_state.has_active_cars():
            prev_state = cur_state
            cur_state = State.from_executing_commands_for_active_cars(
                prev_state, self.field
            )

            collision_history.extend(
                generate_collision_history_for_current_state(prev_state, cur_state)
            )

        return cur_state, collision_history

    def run_simulation_generator(self) -> Iterator[tuple[State, list[CollisionEvent]]]:
        initial_state, collision_history = (
            self.get_initial_state_and_collision_history()
        )

        yield initial_state, collision_history

        prev_state: State
        cur_state: State = initial_state
        while cur_state.has_active_cars():
            prev_state = cur_state
            cur_state = State.from_executing_commands_for_active_cars(
                prev_state, self.field
            )

            collision_history.extend(
                generate_collision_history_for_current_state(prev_state, cur_state)
            )

            yield cur_state, collision_history
