from collections import deque
from dataclasses import dataclass
from typing import Self

from .collision_rubble import CollisionRubble
from .enums import Command
from .field import Field
from .position import Position
from .position_vector import PositionVector


@dataclass
class State:
    id: int

    car_position_map: dict[str, PositionVector]
    active_car_command_map: dict[str, deque[Command]]

    collision_rubble_map: dict[Position, CollisionRubble]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, State):
            return NotImplemented

        return (
            self.id == other.id
            and self.car_position_map == other.car_position_map
            and self.active_car_command_map == other.active_car_command_map
            and self.collision_rubble_map == other.collision_rubble_map
        )

    def has_active_cars(self) -> bool:
        return len(self.active_car_command_map) > 0

    @staticmethod
    def get_rubbles_and_disabled_cars_from_mutual_collisions(
        car_position_map: dict[str, PositionVector]
    ) -> tuple[dict[Position, CollisionRubble], set[str]]:

        collision_map: dict[Position, set[str]] = {}
        for car, position_vector in car_position_map.items():
            position = position_vector.position

            if position not in collision_map:
                collision_map[position] = set()

            collision_map[position].add(car)

        collision_rubble_map: dict[Position, CollisionRubble] = {}
        disabled_car_set: set[str] = set()
        for position, car_set in collision_map.items():
            if len(car_set) == 1:
                continue

            collision_rubble = CollisionRubble(car_set)
            collision_rubble_map[position] = collision_rubble

            disabled_car_set.update(car_set)

        return collision_rubble_map, disabled_car_set

    @staticmethod
    def get_updated_rubbles_and_disabled_cars_from_collisions_with_existing_rubbles(
        collision_rubble_map: dict[Position, CollisionRubble],
        car_position_map: dict[str, PositionVector],
    ) -> tuple[dict[Position, CollisionRubble], set[str]]:

        position_disabled_cars_map: dict[Position, list[str]] = {}
        disabled_cars_set: set[str] = set()

        for car, position_vector in car_position_map.items():
            position = position_vector.position

            if position in collision_rubble_map:
                if position not in position_disabled_cars_map:
                    position_disabled_cars_map[position] = []

                position_disabled_cars_map[position].append(car)
                disabled_cars_set.add(car)

        updated_collision_rubble_map: dict[Position, CollisionRubble] = {}
        for position, collision_rubble in collision_rubble_map.items():

            if position in position_disabled_cars_map:
                disabled_cars = position_disabled_cars_map[position]
                updated_collision_rubble = CollisionRubble.from_adding_cars_to_rubble(
                    collision_rubble, disabled_cars
                )
                updated_collision_rubble_map[position] = updated_collision_rubble
            else:
                updated_collision_rubble_map[position] = collision_rubble

        return updated_collision_rubble_map, disabled_cars_set

    @staticmethod
    def get_tentative_maps_after_executing_commands_for_active_cars(
        car_position_map: dict[str, PositionVector],
        car_command_map: dict[str, deque[Command]],
        field: Field,
    ) -> tuple[dict[str, PositionVector], dict[str, deque[Command]]]:

        tentative_car_position_map: dict[str, PositionVector] = {}
        tentative_car_command_map: dict[str, deque[Command]] = {}

        for car, position in car_position_map.items():
            if car in car_command_map:
                continue

            tentative_car_position_map[car] = position

        for car, command_queue in car_command_map.items():
            cur_position = car_position_map[car]
            next_command_queue = command_queue.copy()

            next_command = next_command_queue.popleft()

            next_position = PositionVector.from_applying_command_to_position_on_field(
                next_command, cur_position, field
            )

            tentative_car_position_map[car] = next_position

            if len(next_command_queue):
                tentative_car_command_map[car] = next_command_queue

        return tentative_car_position_map, tentative_car_command_map

    @classmethod
    def from_initial_simulator_config(
        cls,
        init_car_position_map: dict[str, PositionVector],
        init_car_command_map: dict[str, deque[Command]],
    ) -> Self:

        collision_rubble_map, disabled_car_set = (
            State.get_rubbles_and_disabled_cars_from_mutual_collisions(
                init_car_position_map
            )
        )

        car_position_map: dict[str, PositionVector] = {}
        active_car_command_map: dict[str, deque[Command]] = {}

        for car, position_vector in init_car_position_map.items():
            if car in disabled_car_set:
                continue

            car_position_map[car] = position_vector

            command_queue = init_car_command_map[car]
            if len(command_queue):
                active_car_command_map[car] = init_car_command_map[car]

        return cls(
            id=0,
            car_position_map=car_position_map,
            active_car_command_map=active_car_command_map,
            collision_rubble_map=collision_rubble_map,
        )

    @classmethod
    def from_executing_commands_for_active_cars(
        cls, cur_state: Self, field: Field
    ) -> Self:
        tentative_position_map, tentative_command_map = (
            State.get_tentative_maps_after_executing_commands_for_active_cars(
                cur_state.car_position_map, cur_state.active_car_command_map, field
            )
        )

        # check for cars colliding with existing rubbles
        collision_rubble_map_1, disabled_cars_set = (
            State.get_updated_rubbles_and_disabled_cars_from_collisions_with_existing_rubbles(
                cur_state.collision_rubble_map, tentative_position_map
            )
        )

        # remove cars disabled from collisions with existing rubbles from tentative maps
        for car in disabled_cars_set:
            del tentative_position_map[car]

            if car in tentative_command_map:
                del tentative_command_map[car]

        # check for cars colliding with eachother
        collision_rubble_map_2, disabled_cars_set = (
            State.get_rubbles_and_disabled_cars_from_mutual_collisions(
                tentative_position_map
            )
        )

        # remove cars disabled from collisions with eachother
        for car in disabled_cars_set:
            del tentative_position_map[car]

            if car in tentative_command_map:
                del tentative_command_map[car]

        return cls(
            id=cur_state.id + 1,
            car_position_map=tentative_position_map,
            active_car_command_map=tentative_command_map,
            collision_rubble_map=collision_rubble_map_1 | collision_rubble_map_2,
        )
