from rich import box
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table

from .constants import (
    BOOM_EMOJI,
    DOWN_ARROW_EMOJI,
    LEFT_ARROW_EMOJI,
    RIGHT_ARROW_EMOJI,
    SIMULATION_ONSET_RENDER_TITLE,
    SIMULATION_RENDER_TITLE,
    UP_ARROW_EMOJI,
)
from .enums import Direction
from .field import Field
from .position import Position
from .state import State


def get_computer_coordinates_from_math_coordinates(
    position: Position, field: Field
) -> tuple[int, int]:
    math_x, math_y = position.x, position.y

    x = field.height - 1 - math_y
    y = math_x

    return x, y


def generate_state_representation_on_field(
    state: State, field: Field
) -> list[list[str | Markdown]]:
    field_repr: list[list[str | Markdown]] = [
        [""] * field.width for _ in range(field.height)
    ]

    for position in state.collision_rubble_map.keys():
        x, y = get_computer_coordinates_from_math_coordinates(position, field)
        field_repr[x][y] = BOOM_EMOJI

    for position_vector in state.car_position_map.values():
        position = position_vector.position
        direction = position_vector.direction

        x, y = get_computer_coordinates_from_math_coordinates(position, field)

        if direction == Direction.NORTH:
            field_repr[x][y] = UP_ARROW_EMOJI
        elif direction == Direction.EAST:
            field_repr[x][y] = RIGHT_ARROW_EMOJI
        elif direction == Direction.SOUTH:
            field_repr[x][y] = DOWN_ARROW_EMOJI
        else:
            field_repr[x][y] = LEFT_ARROW_EMOJI

    return field_repr


def render_state_on_field(state: State, field: Field) -> None:

    if state.id == 0:
        title = SIMULATION_ONSET_RENDER_TITLE.format(field)
    else:
        title = SIMULATION_RENDER_TITLE.format(field, state.id)

    table = Table(
        title=title,
        show_lines=True,
        show_header=False,
        show_footer=False,
        row_styles=["dim", ""],
        box=box.ROUNDED,
    )

    field_repr = generate_state_representation_on_field(state, field)

    # add a column for column indices
    table.add_column("", justify="center")

    for _ in range(field.width):
        table.add_column("", justify="center")

    for idx, row_repr in enumerate(field_repr):
        row_idx = str(field.height - 1 - idx)
        row_vals = [row_idx] + row_repr
        table.add_row(*row_vals)

    # last row containing row indices
    table.add_row(*([""] + list(map(str, range(field.width)))))

    console = Console()
    console.print(table)
